from pathlib import Path as PathWrapper
import tempfile, shutil
import os, stat
from typing import Optional


class FsysNode:
    def __init__(self, path : str):
        self._path_wrapper : PathWrapper = PathWrapper(path)
        if not (self.is_dir() or self.is_file()):
            raise FileNotFoundError(f'Path {path} is not a file/folder')

    def get_zip(self) -> bytes:
        with tempfile.TemporaryDirectory() as write_dir:
            zip_base_path = os.path.join(write_dir, 'zipfile')
            args_dir = {
                'base_name': zip_base_path,
                'format': 'zip',
            }
            if self.is_file():
                args_dir['root_dir'] = self.get_dirpath()
                args_dir['base_dir'] = self.get_name()

            if self.is_dir():
                args_dir['root_dir'] = self.get_path()

            shutil.make_archive(**args_dir)
            with open(f'{zip_base_path}.zip', 'rb') as file:
                zip_bytes = file.read()

        return zip_bytes

    def get_subfile_paths(self) -> list[str]:
        subfile_paths = []
        for root, dirs, files in os.walk(self.get_path()):
            for file in files:
                fpath = os.path.join(root, file)
                subfile_paths.append(fpath)
        return subfile_paths

    # -------------------------------------------
    # path and naming


    def get_path(self) -> str:
        return str(self._path_wrapper.absolute())

    def get_dirpath(self) -> str:
        return str(self._path_wrapper.parent.absolute())

    def get_name(self) -> str:
        return self._path_wrapper.name

    def get_suffix(self) -> Optional[str]:
        parts = self.get_name().split('.')
        if len(parts) == 1:
            return None
        else:
            return parts[-1]

    # -------------------------------------------
    # attributes

    def is_file(self) -> bool:
        return self._path_wrapper.is_file()

    def is_dir(self) -> bool:
        return self._path_wrapper.is_dir()

    def is_hidden(self) -> bool:
        if os.name == 'posix':
            return os.path.basename(self.get_path()).startswith('.')
        elif os.name == 'nt':
            return bool(os.stat(self.get_path()).st_file_attributes & stat.FILE_ATTRIBUTE_HIDDEN)
        else:
            raise NotImplementedError(f'Unsupported OS: {os.name}, {FsysNode.is_hidden.__name__} is only supported '
                                      f'on Windows and Unix systems')

    def get_last_modified_epochtime(self) -> float:
        return os.path.getmtime(self.get_path())

    def get_size_in_MB(self) -> float:
        return os.path.getsize(self.get_path()) / (1024 * 1024)

if __name__ == "__main__":
    a = FsysNode(path='/home/daniel/sciai/')
    a.get_subfile_paths()


import linecache
import os
import time
import traceback
import unittest
from typing import Optional
from unittest import TestCase, TestResult
import logging
from .case import CaseReport, CaseStatus, UnitTestCase

# ---------------------------------------------------------

class SuiteRunResult(TestResult):
    test_spaces = 50
    status_spaces = 10
    runtime_space = 10

    def __init__(self, logger : logging.Logger, testsuite_name: str, manual_mode : bool = False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logger
        self.case_reports : list[CaseReport] = []
        self.start_times : dict[str, float] = {}
        self.is_manual : bool = manual_mode
        self.print_header(f'  Test suite: \"{testsuite_name}\"  ')

    def log(self, msg : str, level : int = logging.INFO):
        self.logger.log(msg=msg, level=level)

    def stopTestRun(self):
        super().stopTestRun()
        self.print_summary()

    def startTest(self, case : UnitTestCase):
        if self.is_manual:
            case.set_is_manual()
        self.log(msg=f'------> {case.get_name()[:self.test_spaces]} ', level=logging.INFO)
        self.start_times[case.id()] = time.time()
        super().startTest(case)

    # ---------------------------------------------------------
    # case logging

    # noinspection PyTypeChecker
    def addSuccess(self, case : UnitTestCase):
        super().addSuccess(case)
        self.make_report(case, CaseStatus.SUCCESS)

    # noinspection PyTypeChecker
    def addError(self, case : UnitTestCase, err):
        super().addError(case, err)
        self.make_report(case, CaseStatus.ERROR, err)

    # noinspection PyTypeChecker
    def addFailure(self, case : UnitTestCase, err):
        super().addFailure(case, err)
        self.make_report(case, CaseStatus.FAIL, err)

    # noinspection PyTypeChecker
    def addSkip(self, case : UnitTestCase, reason):
        super().addSkip(case, reason)
        self.make_report(case, CaseStatus.SKIPPED)

    def make_report(self, case : UnitTestCase, status: CaseStatus, err : Optional[tuple] = None):
        if isinstance(case, UnitTestCase):
            report = CaseReport(name=case.get_name(), status=status, runtime=self.get_runtime(case))
            self.case_reports.append(report)

            conditional_err_msg = f'\n{self.get_err_details(err)}' if err else ''
            finish_log_msg = f'Status: {status.value}{conditional_err_msg}\n'
            self.log(msg=finish_log_msg, level=status.get_log_level())
        else:
            raise ValueError(f'Expected CustomTestCase, got {type(case)}')


    def get_runtime(self, test : TestCase) -> Optional[float]:
        test_id = test.id()
        if test_id in self.start_times:
            time_in_sec =  time.time() - self.start_times[test_id]
            return round(time_in_sec, 3)
        else:
            self.log(msg=f'Couldnt find start time of test {test_id}. Current start_times : {self.start_times}',
                     level=logging.ERROR)

    @staticmethod
    def get_err_details(err) -> str:
        err_class, err_instance, err_traceback = err
        tb_list = traceback.extract_tb(err_traceback)

        def is_relevant(tb):
            not_unittest = not os.path.dirname(unittest.__file__) in tb.filename
            not_custom_unittest = not os.path.dirname(__file__) in tb.filename
            return not_unittest and not_custom_unittest

        user_tb = [tb for tb in tb_list if is_relevant(tb)]

        result = ''
        relevant_tb = user_tb if not len(user_tb) == 0 else tb_list
        for frame in relevant_tb:
            file_path = frame.filename
            line_number = frame.lineno
            tb_str = (f'File "{file_path}", line {line_number}, in {frame.name}\n'
                      f'    {linecache.getline(file_path, line_number).strip()}')
            result += f'{err_class.__name__}: {err_instance}\n{tb_str}'
        return result


    # ---------------------------------------------------------
    # summary logging

    def print_summary(self):
        self.print_header(msg=' Summary ', seperator='-')
        for case in self.case_reports:
            level = case.status.get_log_level()
            name_msg = f'{case.name[:self.test_spaces - 4]:<{self.test_spaces}}'
            status_msg = f'{case.status.value:<{self.status_spaces}}'
            runtime_str = f'{case.runtime_sec}s'
            runtime_msg = f'{runtime_str:^{self.runtime_space}}'

            self.log(msg=f'{name_msg}{status_msg}{runtime_msg}', level=level)
        self.log(self.get_final_status())
        self.print_header(msg='')


    def print_header(self, msg: str, seperator : str = '='):
        total_len = self.test_spaces + self.status_spaces
        total_len += self.runtime_space
        line_len = max(total_len- len(msg), 0)
        lines = seperator * int(line_len / 2.)
        self.log(msg=f'{lines}{msg}{lines}')


    def get_final_status(self) -> str:
        num_total = self.testsRun
        num_unsuccessful = len(self.errors)+ len(self.failures)

        RED = '\033[91m'
        GREEN = '\033[92m'
        RESET = '\033[0m'
        CHECKMARK = '✓'
        CROSS = '❌'

        if num_unsuccessful == 0:
            final_status = f"{GREEN}\n{CHECKMARK} {num_total}/{num_total} tests ran successfully!{RESET}"
        else:
            final_status = f"{RED}\n{CROSS} {num_unsuccessful}/{num_total} tests had errors or failures!{RESET}"

        return final_status




import os
import sys
import time
from datetime import datetime
from typing import Callable

def get_timestamp(time_only : bool = False):
    if time_only:
        return datetime.now().strftime("%H:%M:%S")
    else:
        return datetime.now().strftime("%Y-%m-%d_%H%M")


def mute(func : Callable) -> Callable:
    def muted_func(*args, **kwargs):
        stdout, stderr = sys.stdout, sys.stderr
        sys.stdout = open(os.devnull, 'w')
        sys.stderr = open(os.devnull, 'w')
        result = func(*args, **kwargs)
        sys.stdout, sys.stderr = stdout, stderr
        return result
    return muted_func


def log_execution(func):
    def wrapper(*args, **kwargs):
        start_time = time.time()
        timestamp_start = get_timestamp(time_only=True)
        print(f"-[{timestamp_start}]: Launched \"{func.__name__}\" ")

        result = func(*args, **kwargs)

        end_time = time.time()
        timestamp_end = get_timestamp(time_only=True)
        elapsed_time_ms = (end_time - start_time) * 1000
        print(f"-[{timestamp_end}]: Finished \"{func.__name__}\" (Time taken: {elapsed_time_ms:.2f} ms)")

        return result

    return wrapper



def to_sci_notation(val : str | float | int) -> str:
    try:
        val = float(val)
        display_val = f'{val:.1e}'
    except:
        display_val = val
    return  display_val



if __name__ == "__main__":
    @log_execution
    def some_nonsense():
        print('Hello world')

    some_nonsense()


import builtins

from .profiler import Profiler



class ProfiledImportScope:
    def __init__(self):
        self._original_importer = builtins.__import__
        self.profiler : Profiler = Profiler()
        self.is_in_stack : bool = False

    # noinspection PyShadowingBuiltins
    def _profiled_import(self, name, globals=None, locals=None, fromlist=(), level=0):
        if level == 0 and not self.is_in_stack:
            self.is_in_stack = True
            with self.profiler.tracked_scope(name=name):
                result = self._original_importer(name, globals, locals, fromlist, level=level)
            self.is_in_stack = False
        else:
            result = self._original_importer(name, globals, locals, fromlist, level=level)
        return result

    def __enter__(self):
        builtins.__import__ = self._profiled_import

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(self.profiler.scope_report(section_name=f'Library', print_average_times=False, print_num_calls=False))
        builtins.__import__ = self._original_importer




from __future__ import annotations

import logging
import unittest
from enum import Enum

# ---------------------------------------------------------


class UnitTestCase(unittest.TestCase):
    def __init__(self, *args):
        super().__init__(*args)
        self.is_manual_mode : bool = False

    def set_is_manual(self):
        self.is_manual_mode = True

    def get_is_manual(self):
        return self.is_manual_mode

    def get_name(self) -> str:
        full_test_name = self.id()
        parts = full_test_name.split('.')
        last_parts = parts[-2:]
        test_name = '.'.join(last_parts)
        return test_name


class CaseStatus(Enum):
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"
    FAIL = "FAIL"
    SKIPPED = "SKIPPED"

    def get_log_level(self) -> int:
        status_to_logging = {
            CaseStatus.SUCCESS: logging.INFO,
            CaseStatus.ERROR: logging.CRITICAL,
            CaseStatus.FAIL: logging.ERROR,
            CaseStatus.SKIPPED: logging.INFO
        }
        return status_to_logging[self]

class CaseReport:
    def __init__(self, name : str, status : CaseStatus, runtime : float):
        self.runtime_sec : float = runtime
        self.name : str = name
        self.status : CaseStatus = status





from .factory import Loggable, LoggerFactory
from .modifier import LoggingModifier
from .tools import get_timestamp, mute, to_sci_notation


import logging
class LogLevel:
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    WARNING = logging.WARNING
    ERROR = logging.ERROR
    CRITICAL = logging.CRITICAL


import linecache
import os
import traceback
import types
from typing import Literal

Color = Literal['BLACK', 'RED', 'GREEN', 'YELLOW', 'BLUE', 'MAGENTA', 'CYAN', 'WHITE']

# ---------------------------------------------------------


class LoggingModifier:
    @staticmethod
    def add_color(msg : str, color : Color) -> str:
        color_code_map = {
            'BLACK': '\033[30m',
            'RED': '\033[31m',
            'GREEN': '\033[32m',
            'YELLOW': '\033[33m',
            'BLUE': '\033[34m',
            'MAGENTA': '\033[35m',
            'CYAN': '\033[36m',
            'WHITE': '\033[37m'
        }

        color_code = color_code_map[color]
        return f"{color_code}{msg}\033[0m"


    @staticmethod
    def get_limited_stacktrace(err: Exception, excluded_modules : list[types.ModuleType]) -> str:
        err_class, err_instance, err_traceback = err.__class__, err, err.__traceback__
        tb_list = traceback.extract_tb(err_traceback)
        def is_relevant(tb):
            in_exclued_module = any([os.path.dirname(module.__file__) in tb.filename for module in excluded_modules])
            return not in_exclued_module

        err_msg = ''
        relevant_tb = [tb for tb in tb_list if is_relevant(tb)]
        if relevant_tb:
            err_msg = "\nEncountered error during training routine. Relevant stacktrace:"
            for frame in relevant_tb:
                file_path = frame.filename
                line_number = frame.lineno
                tb_str = (f'File "{file_path}", line {line_number}, in {frame.name}\n'
                          f'    {linecache.getline(file_path, line_number).strip()}')
                err_msg += f'\n{err_class.__name__}: {err_instance}\n{tb_str}'

        return err_msg



from __future__ import annotations

import atexit
import tempfile
import time
import webbrowser

from tabulate import tabulate


# ----------------------------------------

class Profiler:
    def __init__(self, print_on_exit : bool = False):
        self.print_on_exit : bool = print_on_exit
        self.scopes : dict[str, TrackedScope] = {}

        if self.print_on_exit:
            def print_report():
                print(f'----> Profiler scope report\n')
                print(self.scope_report())
            atexit.register(print_report)

    def scope_report(self, section_name : str = f'Routine', print_average_times=True, print_num_calls=True) -> str:
        headers = [section_name, "Total Time (s)"]
        if print_average_times:
            headers.append("Average Time (s)")
        if print_num_calls:
            headers.append("Calls")

        table = []
        for name, scope in self.scopes.items():
            row = [name, f"{scope.total_time:.6f}"]
            if print_average_times:
                row.append(f"{scope.average_time:.6f}")
            if print_num_calls:
                row.append(scope.num_calls)
            table.append(row)

        return tabulate(table, headers=headers, tablefmt="psql")

    def show_call_graphs(self):
        for sc in self.scopes.values():
            sc.display()

    def tracked_scope(self, name : str) -> TrackedScope:
        if name in self.scopes:
            ts = self.scopes[name]
        else:
            ts = TrackedScope()
            self.scopes[name] = ts
        return ts


class TrackedScope:
    def __init__(self):
        tmp_fpath = tempfile.mktemp(suffix='.png')
        self.execution_times : list[float] = []
        self.fpath = tmp_fpath

    def __enter__(self):
        self.start_time = time.time()

    def __exit__(self, *args):
        end_time = time.time()
        elapsed = end_time - self.start_time
        self.execution_times.append(elapsed)

    def display(self):
        webbrowser.open('file://' + self.fpath)

    @property
    def num_calls(self):
        return len(self.execution_times)

    @property
    def total_time(self):
        return sum(self.execution_times)

    @property
    def average_time(self):
        return self.total_time / self.num_calls


if __name__ == "__main__":
    pass



import inspect
import linecache
import logging
import os
import tracemalloc
import unittest
import unittest.mock
import warnings
from logging import Logger
from typing import Optional, Callable
from unittest import TestSuite

from holytools.logging import LoggerFactory
from .case import UnitTestCase
from .result import SuiteRunResult


# ---------------------------------------------------------

class Unittest(UnitTestCase):
    _logger : Logger = None

    @classmethod
    def execute_all(cls, manual_mode : bool = True):
        suite = unittest.TestLoader().loadTestsFromTestCase(cls)
        runner = Runner(logger=cls.get_logger(), is_manual=manual_mode, test_name=cls.__name__)
        results = runner.run(testsuite=suite)
        results.print_summary()

        return results

    @classmethod
    def get_logger(cls) -> Logger:
        if not cls._logger:
            cls._logger = LoggerFactory.get_logger(include_location=False, include_timestamp=False, name=cls.__name__)
        return cls._logger

    @classmethod
    def log(cls, msg : str, level : int = logging.INFO):
        cls.get_logger().log(msg=f'{msg}', level=level)

    def warning(self, msg : str, *args, **kwargs):
        kwargs['level'] = logging.WARNING
        self._logger.log(msg=msg, *args, **kwargs)

    def error(self, msg : str, *args, **kwargs):
        kwargs['level'] = logging.ERROR
        self._logger.log(msg=msg, *args, **kwargs)

    def critical(self, msg : str, *args, **kwargs):
        kwargs['level'] = logging.CRITICAL
        self._logger.log(msg=msg, *args, **kwargs)

    def info(self, msg : str, *args, **kwargs):
        kwargs['level'] = logging.INFO
        self._logger.log(msg=msg, *args, **kwargs)


    # ---------------------------------------------------------
    # assertions

    def assertEqual(self, first : object, second : object, msg : Optional[str] = None):
        if not first == second:
            first_str = str(first).__repr__()
            second_str = str(second).__repr__()
            if msg is None:
                msg = (f'Tested expressions should match:'
                       f'\nFirst : {first_str} ({type(first)})'
                       f'\nSecond: {second_str} ({type(second)})')
            raise AssertionError(msg)

    def assertSame(self, first : object, second : object):
        if isinstance(first, float) and isinstance(second, float):
            self.assertSameFloat(first, second)
        else:
            self.assertEqual(first, second)


    @staticmethod
    def assertSameFloat(first : float, second : float, msg : Optional[str] = None):
        if first != first:
            same_float = second != second
        else:
            same_float = first == second
        if not same_float:
            first_str = str(first).__repr__()
            second_str = str(second).__repr__()
            if msg is None:
                msg = (f'Tested floats should match:'
                       f'\nFirst : {first_str} ({type(first)})'
                       f'\nSecond: {second_str} ({type(second)})')
            raise AssertionError(msg)


    def assertIn(self, member : object, container, msg : Optional[str] = None):
        if not member in container:
            member_str = str(member).__repr__()
            container_str = str(container).__repr__()
            if msg is None:
                msg = f'{member_str} not in {container_str}'
            raise AssertionError(msg)


    def assertIsInstance(self, obj : object, cls : type, msg : Optional[str] = None):
        if not isinstance(obj, cls):
            obj_str = str(obj).__repr__()
            cls_str = str(cls).__repr__()
            if msg is None:
                msg = f'{obj_str} not an instance of {cls_str}'
            raise AssertionError(msg)


    def assertTrue(self, expr : bool, msg : Optional[str] = None):
        if not expr:
            if msg is None:
                msg = f'Tested expression should be true'
            raise AssertionError(msg)


    def assertFalse(self, expr : bool, msg : Optional[str] = None):
        if expr:
            if msg is None:
                msg = f'Tested expression should be false'
            raise AssertionError(msg)



    def assert_recursively_same(self, first : dict, second : dict, msg : Optional[str] = None):
        for key in first:
            first_obj = first[key]
            second_obj = second[key]
            self.assertSame(type(first_obj), type(second_obj))
            if isinstance(first_obj, dict):
                self.assert_recursively_same(first_obj, second_obj, msg=msg)
            elif isinstance(first_obj, list):
                for i in range(len(first_obj)):
                    self.assertSame(first_obj[i], second_obj[i])
            else:
                self.assertSame(first_obj, second_obj)

    @staticmethod
    def patch_module(original: type | Callable, replacement: type | Callable):
        module_path = inspect.getmodule(original).__name__
        qualified_name = original.__qualname__
        frame = inspect.currentframe().f_back
        caller_module = inspect.getmodule(frame)

        try:
            # corresponds to "from [caller_module] import [original]
            _ = getattr(caller_module, qualified_name)
            full_path = f"{caller_module.__name__}.{qualified_name}"
        except Exception:
            # corresponds to import [caller_module].[original]
            full_path = f"{module_path}.{qualified_name}"

        # print(f'Full path = {full_path}')
        def decorator(func):
            return unittest.mock.patch(full_path, replacement)(func)

        return decorator


class Runner(unittest.TextTestRunner):
    def __init__(self, logger : Logger, test_name : str, is_manual : bool = False):
        super().__init__(resultclass=None)
        self.logger : Logger = logger
        self.manual_mode : bool = is_manual
        self.test_name : str = test_name

    def run(self, testsuite : TestSuite) -> SuiteRunResult:
        tracemalloc.start(25)
        warnings.simplefilter("always", ResourceWarning)

        with warnings.catch_warnings(record=True) as captured_warnings:
            result = SuiteRunResult(logger=self.logger,
                                    testsuite_name=self.test_name,
                                    stream=self.stream,
                                    descriptions=self.descriptions,
                                    verbosity=2,
                                    manual_mode=self.manual_mode)
            testsuite(result)
            result.printErrors()

        for warning in captured_warnings:
            print(f'- Unclosed resources:')
            print(Runner.warning_to_str(warning_msg=warning))

        warnings.simplefilter("ignore", ResourceWarning)
        tracemalloc.stop()

        return result

    @staticmethod
    def warning_to_str(warning_msg: warnings.WarningMessage) -> str:
        tb = tracemalloc.get_object_traceback(warning_msg.source)
        frames = list(tb)
        frames = [f for f in frames if Runner.is_relevant(frame=f)]

        result = ''
        for frame in frames:
            file_path = frame.filename
            line_number = frame.lineno
            result += (f'File "{file_path}", line {line_number}\n'
                      f'    {linecache.getline(file_path, line_number).strip()}\n')
        return result

    @staticmethod
    def is_relevant(frame):
        not_unittest = not os.path.dirname(unittest.__file__) in frame.filename
        not_custom_unittest = not os.path.dirname(__file__) in frame.filename
        return not_unittest and not_custom_unittest



from __future__ import annotations
from typing import Optional
from abc import abstractmethod, ABC

# -------------------------------------------


class TreeNode(ABC):
    @abstractmethod
    def get_name(self) -> str:
        pass

    # noinspection DuplicatedCode
    def get_tree(self, max_depth : Optional[int] = None, max_size : Optional[int] = None, **kwargs) -> Tree:
        # noinspection DuplicatedCode
        def get_subdict(node : TreeNode, depth : int) -> dict:
            nonlocal root_size
            the_dict = {node : {}}
            root_size += 1

            depth_ok = depth <= max_depth if not max_depth is None else True
            size_ok = root_size <= max_size if not max_size is None else True

            if not depth_ok:
                raise ValueError(f'Exceeded max depth of {max_depth}')
            if not size_ok:
                raise ValueError(f'Exceeded max size of {max_size}')

            child_nodes = node.get_child_nodes(**kwargs)
            for child in child_nodes:
                subtree = get_subdict(node=child, depth=depth+1)
                the_dict[node].update(subtree)
            return the_dict

        root_size = 0
        return Tree(get_subdict(node=self, depth=0))

    # -------------------------------------------
    # descendants

    def get_subnodes(self, *args, **kwargs) -> list[TreeNode]:
        _, __ = args, kwargs
        subnodes = []
        for child in self.get_child_nodes():
            subnodes.append(child)
            subnodes += child.get_subnodes()
        return subnodes

    @abstractmethod
    def get_child_nodes(self, *args, **kwargs) -> list[TreeNode]:
        pass

    # -------------------------------------------
    # ancestors

    def get_ancestors(self) -> list[TreeNode]:
        current = self
        ancestors = []
        while current.get_parent():
            ancestors.append(current.get_parent())
            current = current.get_parent()
        return ancestors

    @abstractmethod
    def get_parent(self) -> Optional[TreeNode]:
        pass

    def get_root(self) -> TreeNode:
        current = self
        while current.get_parent():
            current = current.get_parent()
        return current

    def __str__(self):
        return self.get_name()



class Tree(dict[TreeNode, dict]):
    def as_str(self) -> str:
        return nested_dict_as_str(nested_dict=self)

    def get_size(self) -> int:
        return get_total_elements(nested_dict=self)

    @classmethod
    def join_trees(cls, root : TreeNode, subtrees : list[Tree]):
        the_dict = { root : {}}
        sub_dict = the_dict[root]
        for subtree in subtrees:
            sub_dict.update(subtree)
        return Tree(the_dict)


def nested_dict_as_str(nested_dict: dict, prefix='') -> str:
    output = ''
    for index, (key, value) in enumerate(nested_dict.items()):
        is_last = index == len(nested_dict) - 1
        new_prefix = prefix + ('    ' if is_last else '│   ')
        connector = '└── ' if is_last else '├── '
        output += f'{prefix}{connector}{key}\n'
        output += nested_dict_as_str(nested_dict=value, prefix = new_prefix)
    return output

def get_total_elements(nested_dict : dict) -> int:
    count = 0
    for key, value in nested_dict.items():
        count += 1
        if isinstance(value, dict):
            count += get_total_elements(value)
    return count





from typing import Callable
import inspect
from holytools.logging import Loggable, LogLevel
# ---------------------------------------------------------


class CLI(Loggable):
    _exit_str = 'q'

    def __init__(self, cls : type, description : str = ''):
        super().__init__()
        self._cls : type = cls
        self._desc : str = description

        self._log_header()
        self.obj : object = self._create_object()
        self.methods_dict : dict[int, Callable] = self._get_methods_dict(obj=self.obj)


    def _log_header(self):
        header_size = 40
        cls_name = self._cls.__name__
        hash_count = int(max(header_size-len(cls_name), 0)/2)
        hashes = '-' * hash_count
        self.log(f'{hashes} {cls_name.upper()} CLI {hashes}')
        desc_str = self._desc if self._desc else 'No description found'
        self.log(f"Description: {desc_str} \n")


    def _create_object(self):
        self.log(f'Initializing object {self._cls.__name__}:')
        try:
            init_kwargs = self._get_args_dict(mthd=self._cls.__init__)
            return self._cls(**init_kwargs)
        except Exception as e:
            self.log(f"Error initializing {self._cls.__name__}: {e}. Please try again\n")
            return self._create_object()



    def _get_methods_dict(self, obj) -> dict[int, Callable]:
        public_methods_names = self._get_public_method_names()
        return {i + 1: getattr(obj,name) for i, name in enumerate(public_methods_names)}


    def _get_public_method_names(self) -> list[str]:
        is_public_callable = lambda attr: callable(getattr(self.obj, attr)) and not attr.startswith("_")
        public_method_names = [method for method in dir(self.obj) if is_public_callable(method)]
        return public_method_names

    # ---------------------------------------------------------
    # loop

    def command_loop(self):
        while True:
            self.print_info()
            user_input = input()
            if user_input == self._exit_str:
                break

            if not user_input.isdigit() or int(user_input) not in self.methods_dict:
                self.log("Please enter a valid number.")
                continue

            try:
                result = self._handle_mthd_call(int(user_input))
                msg = f"Result : {result}"
                level = LogLevel.INFO
            except Exception as e:
                msg = f"Error: {e}"
                level = LogLevel.WARNING
            self.log(msg=msg, level=level)


    def print_info(self):
        text = f"\nChoose a method by entering its number or '{self._exit_str}' to quit): "
        for index, method in self.methods_dict.items():
            text += f"\n{index}: {method.__name__}"
        self.log(msg=text)


    def _handle_mthd_call(self, index : int):
        mthd = self.methods_dict[index]
        args_dict = self._get_args_dict(mthd=mthd)
        result = mthd(**args_dict)
        return result


    def _get_args_dict(self, mthd: Callable) -> dict:
        args_dict = {}
        spec = inspect.getfullargspec(mthd)
        annotations = spec.annotations
        for argname in spec.args[1:]:
            typ = annotations.get(argname, str)
            if isinstance(typ, str):
                typ = type(typ)
            user_input = input(f"Enter value for \"{mthd.__name__}\" argument \"{argname}\" ({typ.__name__}): ")
            args_dict[argname] = self._get_value(user_input=user_input, arg_type=typ, arg_name=argname)
        return args_dict


    @staticmethod
    def _get_value(user_input : str, arg_type : type, arg_name : str):
        if arg_type == bool:
            if user_input not in ['0', '1']:
                raise ValueError(f"For argument '{arg_name}', please enter '0' for False or '1' for True.")
            val = bool(int(user_input))
        else:
            try:
                val = arg_type(user_input)
            except ValueError:
                raise ValueError(f"Invalid input type for '{arg_name}'. Expected a value of type {arg_type.__name__}.")
        return val



if __name__ == "__main__":
    pass